# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
import os
import sys
from threading import current_thread

from mo_dots import Data, coalesce, listwrap, unwraplist, dict_to_data, is_data, to_data
from mo_future import is_text, STDOUT, utcnow
from mo_imports import delay_import
from mo_kwargs import override

from mo_logs import constants as _constants, exceptions, strings
from mo_logs.exceptions import (
    Except,
    LogItem,
    suppress_exception,
    format_trace,
    WARNING,
    get_stacktrace,
    ERROR
)
from mo_logs.log_usingPrint import StructuredLogger_usingPrint
from mo_logs.strings import CR, indent

STACKTRACE = "\n{trace_text|indent}\n{cause_text}"

StructuredLogger_usingMulti = delay_import("mo_logs.log_usingMulti.StructuredLogger_usingMulti")
StructuredLogger_usingThread = delay_import("mo_logs.log_usingThread.StructuredLogger_usingThread")

startup_read_settings = delay_import("mo_logs.startup.read_settings")

all_log_callers = {}


class Log(object):
    """
    FOR STRUCTURED LOGGING AND EXCEPTION CHAINING
    """
    trace = False
    main_log = StructuredLogger_usingPrint()
    logging_multi = None
    profiler = None  # simple pypy-friendly profiler
    error_mode = False  # prevent error loops
    extra = {}
    static_template = True

    @classmethod
    @override("settings")
    def start(
        cls,
        trace=False,
        cprofile=False,
        constants=None,
        logs=None,
        extra=None,
        app_name=None,
        settings=None,
        static_template=True,
    ):
        """
        RUN ME FIRST TO SETUP THE THREADED LOGGING
        https://fangpenlin.com/posts/2012/08/26/good-logging-practice-in-python/

        :param trace: SHOW MORE DETAILS IN EVERY LOG LINE (default False)
        :param cprofile: True==ENABLE THE C-PROFILER THAT COMES WITH PYTHON (default False)
                         USE THE LONG FORM TO SET THE FILENAME {"enabled": True, "filename": "cprofile.tab"}
        :param constants: UPDATE MODULE CONSTANTS AT STARTUP (PRIMARILY INTENDED TO CHANGE DEBUG STATE)
        :param logs: LIST OF PARAMETERS FOR LOGGER(S)
        :param extra: ADDITIONAL DATA TO BE INCLUDED IN EVERY LOG LINE
        :param app_name: GIVE THIS APP A NAME, AND RETURN A CONTEXT MANAGER
        :param settings: ALL THE ABOVE PARAMETERS
        :return:
        """
        if app_name:
            return LoggingContext(app_name)

        Log.stop()

        cls.settings = settings
        cls.trace = trace
        cls.static_template = static_template

        # ENABLE CPROFILE
        if cprofile is False:
            cprofile = settings.cprofile = Data(enabled=False)
        elif cprofile is True:
            cprofile = settings.cprofile = Data(enabled=True, filename="cprofile.tab")
        if is_data(cprofile) and cprofile.enabled:
            from mo_threads import profiles

            profiles.enable_profilers(settings.cprofile.filename)

        if constants:
            _constants.set(constants)

        logs = coalesce(settings.log, logs)
        if logs:
            cls.logging_multi = StructuredLogger_usingMulti()
            for log in listwrap(logs):
                Log._add_log(Log.new_instance(log))

            from mo_logs.log_usingThread import StructuredLogger_usingThread

            old_log, cls.main_log = (
                cls.main_log,
                StructuredLogger_usingThread(cls.logging_multi),
            )
            old_log.stop()
        cls.extra = extra or {}

    @classmethod
    def stop(cls):
        """
        DECONSTRUCTS ANY LOGGING, AND RETURNS TO DIRECT-TO-stdout LOGGING
        EXECUTING MULUTIPLE TIMES IN A ROW IS SAFE, IT HAS NO NET EFFECT, IT STILL LOGS TO stdout
        :return: NOTHING
        """
        old_log, cls.main_log = cls.main_log, StructuredLogger_usingPrint()
        old_log.stop()
        cls.trace = False
        cls.cprofile = False

    @classmethod
    @override("settings")
    def new_instance(cls, log_type=None, settings=None):
        if settings["class"]:
            from mo_logs.log_usingHandler import StructuredLogger_usingHandler

            return StructuredLogger_usingHandler(settings)

        clazz = _known_loggers.get(log_type.lower())
        if clazz:
            return clazz(settings)
        logger.error("Log type of {config|json} is not recognized", config=settings)

    @classmethod
    def _add_log(cls, log):
        cls.logging_multi.add_log(log)

    @classmethod
    def set_logger(cls, logger):
        if cls.logging_multi:
            cls.logging_multi.add_log(logger)
        else:
            from mo_logs.log_usingThread import StructuredLogger_usingThread

            old_log, cls.main_log = cls.main_log, StructuredLogger_usingThread(logger)
            old_log.stop()

    @classmethod
    def note(cls, template, default_params={}, *, stack_depth=0, static_template=None, **more_params):
        """
        :param template: *string* human readable string with placeholders for parameters
        :param default_params: *dict* parameters to fill in template
        :param stack_depth:  *int* how many calls you want popped off the stack to report the *true* caller
        :param static_template: *bool* if True, then the template is static, and optimization can be done
        :param more_params: *any more parameters (which will overwrite default_params)
        :return:
        """
        timestamp = utcnow()
        if not is_text(template):
            logger.error("logger.info was expecting a string template")

        Log._annotate(
            LogItem(
                severity=exceptions.NOTE,
                template=template,
                params=dict(default_params, **more_params),
                timestamp=timestamp,
            ),
            stack_depth + 1,
            cls.static_template if static_template is None else static_template,
        )

    info = note

    @classmethod
    def alarm(cls, template, default_params={}, *, stack_depth=0, static_template=None, **more_params):
        """
        :param template: *string* human readable string with placeholders for parameters
        :param default_params: *dict* parameters to fill in template
        :param stack_depth:  *int* how many calls you want popped off the stack to report the *true* caller
        :param more_params: more parameters (which will overwrite default_params)
        :return:
        """
        timestamp = utcnow()
        template = ("*" * 80) + CR + indent(template, prefix="** ").strip() + CR + ("*" * 80)
        Log._annotate(
            LogItem(
                severity=exceptions.ALARM,
                template=template,
                params=dict(default_params, **more_params),
                timestamp=timestamp,
            ),
            stack_depth + 1,
            cls.static_template if static_template is None else static_template,
        )

    alert = alarm

    @classmethod
    def warning(
        cls,
        template: str,  # human readable string with placeholders for parameters
        default_params={},  # parameters to fill in template
        cause=None,  # for chaining
        *,
        stack_depth=0,  # how many calls you want popped off the stack to report the *true* caller
        log_severity=WARNING,  # set the logging severity
        exc_info=None,  # used by python logging as the cause
        static_template=None,
        **more_params,  # any more parameters (which will overwrite default_params)
    ):
        if exc_info is True:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            exc_info = Except.wrap(exc_value)
        if not is_text(template):
            logger.error("logger.warning was expecting a string template")
        if "values" in more_params.keys():
            logger.error("Can not handle a logging parameter by name `values`")

        if isinstance(default_params, BaseException):
            cause = default_params
            default_params = {}

        params = to_data(dict(default_params, **more_params))
        cause = unwraplist([Except.wrap(c, stack_depth=2) for c in listwrap(cause or exc_info)])
        trace = exceptions.get_stacktrace(stack_depth + 1)

        e = Except(severity=log_severity, template=template, params=params, cause=cause, trace=trace,)
        Log._annotate(e, stack_depth + 1, cls.static_template if static_template is None else static_template)

    warn = warning

    @classmethod
    def error(
        cls,
        template,  # human readable template
        default_params={},  # parameters for template
        *,
        cause=None,  # pausible cause
        stack_depth=0,
        exc_info=None,  # used by python logging as the cause
        **more_params,
    ):
        """
        raise an exception with a trace for the cause too

        :param template: *string* human readable string with placeholders for parameters
        :param default_params: *dict* parameters to fill in template
        :param cause: *Exception* for chaining
        :param stack_depth:  *int* how many calls you want popped off the stack to report the *true* caller
        :param exc_info: *Exception* alternate to cause (used by other logging libs)
        :param more_params: *any more parameters (which will overwrite default_params)
        :return:
        """
        if not is_text(template):
            # sys.stderr.write(str("logger.error was expecting a string template"))
            logger.error("logger.error was expecting a string template")
        if "values" in more_params.keys():
            logger.error("Can not handle a logging parameter by name `values`")
        if exc_info is True:
            exc_info = Except.wrap(sys.exc_info()[1])

        if isinstance(default_params, BaseException):
            cause = default_params
            default_params = {}

        params = to_data(dict(default_params, **more_params))
        cause = unwraplist([Except.wrap(c, stack_depth=2) for c in listwrap(cause or exc_info)])
        trace = exceptions.get_stacktrace(stack_depth + 1)

        e = Except(severity=exceptions.ERROR, template=template, params=params, cause=cause, trace=trace,)
        raise_from_none(e)

    @classmethod
    def _annotate(cls, item, stack_depth, static_template):
        """
        :param item:  A LogItem THE TYPE OF MESSAGE
        :param stack_depth: FOR TRACKING WHAT LINE THIS CAME FROM
        :return:
        """
        given_template = item.template
        given_template = strings.limit(given_template, 10000)
        param_template = "".join(
            f"{text}{{params.{code}}}" if code else text for text, code in strings.parse_template(given_template)
        )

        if isinstance(item, Except):
            param_template = "{severity}: " + param_template + STACKTRACE
            temp = item.__data__()
            temp.trace_text = item.trace_text
            temp.cause_text = item.cause_text
            item = temp
        else:
            item = item.__data__()

        if not param_template.startswith(CR) and CR in param_template:
            param_template = CR + param_template

        if cls.trace:
            item.machine = machine_metadata()
            log_format = item.template = (
                "{machine.name} (pid {machine.pid}) - {timestamp|datetime} -"
                ' {thread.name} - ""{location.file}:{location.line}"" -'
                " ({location.method}) - "
                + param_template
            )
            f = sys._getframe(stack_depth + 1)
            item.location = {
                "line": f.f_lineno,
                "file": f.f_code.co_filename,
                "method": f.f_code.co_name,
            }
            if static_template:
                last_caller_loc = (f.f_code.co_filename, f.f_lineno)
                prev_template = all_log_callers.get(last_caller_loc)
                if prev_template != given_template:
                    if prev_template:
                        raise Except(
                            template="Expecting logger call to be static: was {a|quote} now {b|quote}",
                            params={"a": prev_template, "b": given_template},
                            trace=get_stacktrace(stack_depth + 1),
                        )
                    all_log_callers[last_caller_loc] = given_template
            thread = current_thread()
            item.thread = {"name": thread.name, "id": thread.ident}
        else:
            log_format = param_template
            # log_format = item.template = "{timestamp|datetime} - " + template

        item.params = {**cls.extra, **item.params}
        cls.main_log.write(log_format, item)

    def write(self):
        raise NotImplementedError


logger = Log


def getLogger(*args, **kwargs):
    return logger


class LoggingContext:
    def __init__(self, app_name=None):
        self.app_name = app_name
        self.config = None

    def __enter__(self):
        self.config = config = startup_read_settings()
        from mo_logs import constants

        constants.set(config.constants)
        Log.start(config.debug)
        return config

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_val:
            logger.warning(
                "Problem with {name}! Shutting down.", name=self.app_name, cause=exc_val,
            )
        Log.stop()


def _same_frame(frameA, frameB):
    return (frameA.line, frameA.file) == (frameB.line, frameB.file)


# GET THE MACHINE METADATA
_machine_metadata = None


def machine_metadata():
    global _machine_metadata
    if _machine_metadata:
        return _machine_metadata

    import platform

    _machine_metadata = dict_to_data({
        "pid": os.getpid(),
        "python": platform.python_implementation(),
        "os": (platform.system() + platform.release()).strip(),
        "name": platform.node(),
    })
    return _machine_metadata


def raise_from_none(e):
    raise e from None


def _using_logger(config):
    from mo_logs.log_usingLogger import StructuredLogger_usingLogger

    return StructuredLogger_usingLogger(config)


def _using_file(config):
    from mo_logs.log_usingFile import StructuredLogger_usingFile

    if config.file:
        return StructuredLogger_usingFile(config.file)
    if config.filename:
        return StructuredLogger_usingFile(config.filename)


def _using_console(config):
    from mo_logs.log_usingThread import StructuredLogger_usingThread

    return StructuredLogger_usingThread(StructuredLogger_usingPrint())


def _using_mozlog(config):
    from mo_logs.log_usingMozLog import StructuredLogger_usingMozLog

    return StructuredLogger_usingMozLog(STDOUT, coalesce(config.app_name, config.appname))


def _using_stream(config):
    from mo_logs.log_usingThread import StructuredLogger_usingThread
    from mo_logs.log_usingStream import StructuredLogger_usingStream

    return StructuredLogger_usingThread(StructuredLogger_usingStream(config.stream))


def _using_elasticsearch(config):
    from jx_elasticsearch.log_usingElasticSearch import StructuredLogger_usingElasticSearch

    return StructuredLogger_usingElasticSearch(config)


def _using_email(config):
    from mo_logs.log_usingEmail import StructuredLogger_usingEmail

    return StructuredLogger_usingEmail(config)


def _using_ses(config):
    from mo_logs.log_usingSES import StructuredLogger_usingSES

    return StructuredLogger_usingSES(config)


def _using_nothing(config):
    from mo_logs.log_usingNothing import StructuredLogger

    return StructuredLogger()


_known_loggers = {
    "logger": _using_logger,
    "nothing": _using_nothing,
    "none": _using_nothing,
    "null": _using_nothing,
    "file": _using_file,
    "console": _using_console,
    "mozlog": _using_mozlog,
    "stream": _using_stream,
    "elasticsearch": _using_elasticsearch,
    "email": _using_email,
    "ses": _using_ses,
}


def register_logger(name, factory):
    _known_loggers[name] = factory



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Author: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from math import pi

from mo_testing.fuzzytestcase import FuzzyTestCase

from mo_logs.strings import expand_template


class TestFormatters(FuzzyTestCase):
    def test_upper(self):
        test = expand_template("Hello {{name|upper}}", {"name": "world"})
        self.assertEqual(test, "Hello WORLD")

    def test_places(self):
        test = expand_template("pi = {{pi|round(places=3)}}", {"pi": pi})
        self.assertEqual(test, "pi = 3.14")

    def test_round(self):
        expected = [
            "0.0000000003142",
            "0.000000003142",
            "0.00000003142",
            "0.0000003142",
            "0.000003142",
            "0.00003142",
            "0.0003142",
            "0.003142",
            "0.03142",
            "0.3142",
            "3.142",
            "31.42",
            "314.2",
            "3142",
            "31420",
            "314200",
            "3142000",
            "31420000",
            "314200000",
            "3142000000",
            "31420000000",
            "314200000000",
            "3142000000000",
            "31420000000000",
            "314200000000000",
            "3142000000000000",
            "31420000000000000",
            "314200000000000000",
        ]
        start = -10
        for order in range(start, 18):
            value = pi * (10 ** order)
            test = expand_template("{{value|round(places=4)}}", value={"value": value})
            self.assertEqual(test, expected[order - start])



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Author: Kyle Lahnakoski (kyle@lahnakoski.com)
#
import logging
from unittest import skip

from mo_dots import Data
from mo_future import StringIO
from mo_kwargs import override
from mo_math import randoms
from mo_testing.fuzzytestcase import FuzzyTestCase
from mo_threads import Till, stop_main_thread, start_main_thread
from mo_times import Date

from mo_logs import logger as log, register_logger
from mo_logs.log_usingNothing import StructuredLogger
from mo_logs.strings import expand_template
from tests.utils import add_error_reporting
from tests.utils.udp_listener import UdpListener

UDP_PORT_RANGE = Data(FROM=12200, LENGTH=4000)


@add_error_reporting
class TestLoggers(FuzzyTestCase):
    def setUp(self):
        log.start()

    def tearDown(self):
        log.stop()

    @classmethod
    def setUpClass(cls):
        stop_main_thread()
        start_main_thread()

    @classmethod
    def tearDownClass(cls):
        stop_main_thread()

    def test_logging(self):
        from importlib import reload

        logging.shutdown()
        reload(logging)

        log_stream = StringIO()
        logging.getLogger()
        logging.basicConfig(stream=log_stream, level=logging.INFO)

        log.start(trace=False, settings={"logs": {"log_type": "logger"}})
        log.note("testing")
        while log.main_log.logger.many[0].count < 1:
            Till(seconds=0.1).wait()
        logs = log_stream.getvalue()

        expected = "testing\n"
        self.assertEqual(logs[-len(expected) :], expected)

    def test_graylogger(self):
        port = UDP_PORT_RANGE.FROM + randoms.int(UDP_PORT_RANGE.LENGTH)
        with UdpListener(port) as udp:
            log.start(settings={"logs": {"class": "graypy.GELFUDPHandler", "host": "localhost", "port": port}},)
            log.note("testing {{value}}", value="test")
            message = udp.queue.pop()

        self.assertTrue(message["short_message"].endswith("testing test"))
        self.assertTrue(message["full_message"].endswith("testing test"))
        self.assertTrue(message["file"].endswith("test_loggers.py"))

        print(message)
        self.assertEqual(
            message,
            {
                "_value": "test",
                "_function": "test_graylogger",
                "_process_name": "MainProcess",
                "facility": "mo-logs",
                "level": 6,
                "line": 70,  # <-- CAREFUL WHEN REFORMATTING THIS FILE, THIS CAN CHANGE
                "version": "1.0",
                "_thread_name": "MainThread",
            },
        )
        self.assertIsNone(message["_stack_info"])

    def test_graylogger_for_debugging(self):
        port = UDP_PORT_RANGE.FROM + randoms.int(UDP_PORT_RANGE.LENGTH)
        with UdpListener(port) as udp:
            log.main_log = log.new_instance({"class": "graypy.GELFUDPHandler", "host": "localhost", "port": port})
            log.note("testing {{value}}", value="test")
            message = udp.queue.pop()

        self.assertTrue(message["short_message"].endswith("testing test"))
        self.assertTrue(message["full_message"].endswith("testing test"))
        self.assertTrue(message["file"].endswith("test_loggers.py"))

        print(message)
        self.assertEqual(
            message,
            {
                "_value": "test",
                "_function": "test_graylogger_for_debugging",
                "_process_name": "MainProcess",
                "facility": "mo-logs",
                "level": 6,
                "version": "1.0",
                "_thread_name": "MainThread",
            },
        )
        self.assertIsNone(message["_stack_info"])

    def test_extras(self):
        port = UDP_PORT_RANGE.FROM + randoms.int(UDP_PORT_RANGE.LENGTH)
        with UdpListener(port) as udp:
            log.start(
                settings={
                    "logs": {"class": "graypy.GELFUDPHandler", "host": "localhost", "port": port,},
                    "extra": {"some_name": {"v": "some_value"}},
                },
            )
            log.note("testing {{value}}", value="test")
            message = udp.queue.pop()

        self.assertIsInstance(message, dict)
        self.assertEqual(message, {"_some_name.v": "some_value"})
        self.assertTrue(message["file"].endswith("test_loggers.py"))

    def test_graylogger_exception(self):
        port = randoms.int(UDP_PORT_RANGE.FROM + UDP_PORT_RANGE.LENGTH)
        with UdpListener(port) as udp:
            log.start(settings={"logs": {"class": "graypy.GELFUDPHandler", "host": "localhost", "port": port,}},)
            log.warning("testing {{value}}", value="test")
            message = udp.queue.pop()

        self.assertNotIn('test_loggers.py"', message["short_message"])
        self.assertIn('test_loggers.py"', message["full_message"])
        self.assertIsInstance(message["_stack_info"], str)

    def test_exc_info(self):
        log.start(logs={"log_type": "array"})
        try:
            raise Exception("kyle")
        except Exception:
            log.warning("report", exc_info=True)
            Till(seconds=0.2).wait()

        template, detail = log.main_log.logger.many[0].lines[0]
        self.assertIn("kyle", detail.cause.template)
        self.assertIn("test_loggers.py", detail.cause.trace[0].file)

    def test_exc_info2(self):
        log.start(logs={"log_type": "array"})
        try:
            raise Exception("kyle")
        except Exception as cause:
            log.warning("report", exc_info=cause)
            Till(seconds=0.2).wait()

        template, detail = log.main_log.logger.many[0].lines[0]
        self.assertIn("kyle", detail.cause.template)
        self.assertIn("test_loggers.py", detail.cause.trace[0].file)

    def test_single_braces(self):
        log.trace = True
        logger = log.main_log = LogUsingLines()
        log.note("this is a {test}", test="test")
        # EG: 'kyle-win10 (pid 14900) - 2023-05-29 12:36:09.304071 - MainThread - "C:\\Users\\kyle\\code\\mo-logs\\tests\\test_loggers.py:162" - (test_simple_logging) - this is a test'
        self.assertIn("(pid ", logger.lines[0])
        self.assertIn(" - MainThread - ", logger.lines[0])
        self.assertIn(" - (test_single_braces) - ", logger.lines[0])
        self.assertIn(" - this is a test", logger.lines[0])
        self.assertIn("test_loggers.py:", logger.lines[0])

    def test_bad_log_call(self):
        log.start(trace=True)
        with self.assertRaises("Expecting logger call to be static"):
            for v in ["1", "2"]:
                log.note(v)

    def test_simple_tet(self):
        log.start(trace=False)
        logger = log.main_log = LogUsingLines()
        log.note("Timer start: get modules' status")
        self.assertEqual("Timer start: get modules' status", logger.lines[0])

    @skip("not handled yet")
    def test_log_bad_template(self):
        log.start(trace=False)
        logger = log.main_log = LogUsingLines()
        log.info("Timer start: get modules status {{")
        self.assertEqual("Timer start: get modules status {{", logger.lines[0])

    @skip("not handled yet")
    def test_log_bad_template2(self):
        log.start(trace=False)
        logger = log.main_log = LogUsingLines()
        log.info("Timer start: get modules status {{}}")
        self.assertEqual("Timer start: get modules status ", logger.lines[0])

    @skip("not handled yet")
    def test_log_bad_template3(self):
        log.start(trace=False)
        logger = log.main_log = LogUsingLines()
        log.info("Timer start: get modules status {{'}}")
        self.assertEqual("Timer start: get modules status {{'}}", logger.lines[0])

    @skip("not handled yet")
    def test_log_bad_template4(self):
        log.start(trace=False)
        logger = log.main_log = LogUsingLines()
        log.info("Timer start: get modules status {'}")
        self.assertEqual("Timer start: get modules' status {{", logger.lines[0])

    def test_static_mode_off(self):
        log.start(static_template=False)
        logger = log.main_log = LogUsingLines()
        for i in range(2):
            log.info(f"line {i}")
        self.assertEqual("line 0", logger.lines[0])
        self.assertEqual("line 1", logger.lines[1])

    def test_hex(self):
        result = expand_template("{value|hex}", {"value":"test"})
        expected = '74657374'
        self.assertEqual(result, expected)

    def test_date(self):
        date = Date("2023-12-10")
        port = randoms.int(UDP_PORT_RANGE.FROM + UDP_PORT_RANGE.LENGTH)
        with UdpListener(port) as udp:
            log.start(settings={"logs": {"class": "graypy.GELFUDPHandler", "host": "localhost", "port": port,}},)
            log.info("date {date2}", date2=date)
            message = udp.queue.pop()

        self.assertEqual('2023-12-10T00:00:00Z', message["_date2"])
        self.assertIn("date 2023-12-10 00:00:00", message["full_message"])

    def test_data(self):
        data = Data(x=1)
        port = randoms.int(UDP_PORT_RANGE.FROM + UDP_PORT_RANGE.LENGTH)
        with UdpListener(port) as udp:
            log.start(settings={"logs": {"class": "graypy.GELFUDPHandler", "host": "localhost", "port": port,}},)
            log.info("data {data}", data=data)
            message = udp.queue.pop()

        self.assertEqual(1, message["_data.x"])
        self.assertIn('data {"x": 1}', message["full_message"])



class LogUsingArray(StructuredLogger):
    @override
    def __init__(self, kwargs=None):
        self.lines = []

    def write(self, template, params):
        self.lines.append((template, params))


class LogUsingLines(StructuredLogger):
    @override
    def __init__(self, kwargs=None):
        self.lines = []

    def write(self, template, params):
        value = expand_template(template, params)
        self.lines.append(value)


register_logger("array", LogUsingArray)



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
import logging

from mo_dots import from_data, dict_to_data
from mo_imports import delay_import
from mo_kwargs import override

from mo_logs import logger, STACKTRACE
from mo_logs.exceptions import FATAL, ERROR, WARNING, ALARM, UNEXPECTED, INFO, NOTE, format_trace
from mo_logs.log_usingNothing import StructuredLogger
from mo_logs.strings import expand_template

Log = delay_import("mo_logs.Log")
NO_ARGS = tuple()


# WRAP PYTHON CLASSIC logger OBJECTS
class StructuredLogger_usingHandler(StructuredLogger):
    @override("settings")
    def __init__(self, settings):
        try:
            Log.trace = True  # ENSURE TRACING IS ON SO DETAILS ARE CAPTURED
        except Exception as cause:
            Log.trace = True
        self.count = 0
        self.handler = make_handler_from_settings(settings)

    def write(self, template, params):
        record = logging.LogRecord(
            name="mo-logs",
            level=_severity_to_level[params.severity],
            pathname=params.location.file,
            lineno=params.location.line,
            msg=expand_template(template.replace(STACKTRACE, ""), params),
            args=NO_ARGS,
            exc_info=None,
            func=params.location.method,
            sinfo=format_trace(params.trace) or None,
        )
        record.thread = params.thread.id
        record.threadName = params.thread.name
        record.process = params.machine.pid

        record.exc_text = expand_template(template, params)
        for k, v in params.params.leaves():
            if v.__class__.__name__ == "Date":
                ms = round(float("0."+v.format("%f")), 3)
                if not ms:
                    ms = ""
                else:
                    ms = ms[1:]
                v = v.format(f"%Y-%m-%dT%H:%M:%S{ms}Z")
            setattr(record, k, v)
        self.handler.handle(record)
        self.count += 1

    def stop(self):
        self.handler.flush()
        self.handler.close()


def make_handler_from_settings(settings):
    assert settings["class"]
    settings.self = None

    settings = dict_to_data({**settings})

    # IMPORT MODULE FOR HANDLER
    path = settings["class"].split(".")
    class_name = path[-1]
    path = ".".join(path[:-1])
    constructor = None
    try:
        temp = __import__(path, globals(), locals(), [class_name], 0)
        constructor = object.__getattribute__(temp, class_name)
    except Exception as cause:
        logger.error("Can not find class {{class}}", {"class": path}, cause=cause)

    # IF WE NEED A FILE, MAKE SURE DIRECTORY EXISTS
    if settings.filename != None:
        from mo_files import File

        f = File(settings.filename)
        if not f.parent.exists:
            f.parent.create()

    settings["class"] = None
    settings["cls"] = None
    settings["log_type"] = None
    settings["settings"] = None
    params = from_data(settings)
    try:
        log_instance = constructor(**params)
        return log_instance
    except Exception as cause:
        logger.error("problem with making handler", cause=cause)


_severity_to_level = {
    FATAL: logging.CRITICAL,
    ERROR: logging.ERROR,
    WARNING: logging.WARNING,
    ALARM: logging.INFO,
    UNEXPECTED: logging.CRITICAL,
    INFO: logging.INFO,
    NOTE: logging.INFO,
}



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Author: Kyle Lahnakoski (kyle@lahnakoski.com)
#
import logging
import sys
import unittest
import zlib
from unittest import skip

from mo_dots import listwrap, wrap, Data, to_data
from mo_dots.objects import DataObject
from mo_json import value2json
from mo_logs.exceptions import ERROR, TOO_DEEP
from mo_testing.fuzzytestcase import FuzzyTestCase
from mo_threads import Till

from mo_logs import Except, logger

try:
    from tests.utils.log_usingQueue import StructuredLogger_usingQueue
except Exception as e:
    from test_logs.utils.log_usingQueue import StructuredLogger_usingQueue


class TestExcept(FuzzyTestCase):
    @classmethod
    def setUpClass(cls):
        logger.start({"trace": False})

    def test_trace_of_simple_raises(self):
        try:
            problem_a()
        except Exception as e:
            f = Except.wrap(e)
            self.assertEqual(f.template, "Exception: expected exception")
            for i, m in enumerate(listwrap(f.trace).method):
                if m == "test_trace_of_simple_raises":
                    self.assertEqual(i, 2)
                    break
            else:
                self.fail("expecting stack to show this method")

    def test_full_trace_exists(self):
        try:
            problem_a2()
        except Exception as e:
            cause = e.cause
            self.assertEqual(cause.template, "Exception: expected exception")

            for i, m in enumerate(listwrap(cause.trace).method):
                if m == "test_full_trace_exists":
                    self.assertEqual(i, 2)
                    break
            else:
                self.fail("expecting stack to show this method")

    def test_bad_log_params(self):
        for call in [logger.info, logger.warning, logger.error]:
            with self.assertRaises("was expecting a string template"):
                call({})

    def test_full_trace_on_wrap(self):
        try:
            problem_b()
        except Exception as e:
            cause = Except.wrap(e)
            self.assertEqual(cause.template, "Exception: expected exception")

            for i, m in enumerate(listwrap(cause.trace).method):
                if m == "test_full_trace_on_wrap":
                    self.assertEqual(i, 1)
                    break
            else:
                self.fail("expecting stack to show this method")

    @skip("not implemented")
    def test_local_variable_capture(self):
        a = {"c": "a", "b": "d"}
        b = {"c": "b"}

        A = '{\n    "b": "d",\n    "c": "a"\n}'
        B = '{"c": "b"}'

        log_queue = StructuredLogger_usingQueue("abba")
        backup_log, logger.main_log = logger.main_log, log_queue

        logger.info("{{a}} and {{b}}")
        self.assertEqual(logger.main_log.pop(), A + " and " + B)

        logger.warning("{{a}} and {{b}}", a=a, b=b)
        self.assertEqual(logger.main_log.pop(), A + " and " + B)

    @skip("not implemented")
    def test_missing_local_variable(self):
        a = {"c": "a", "b": "d"}
        b = {"c": "b"}

        log_queue = StructuredLogger_usingQueue("abba")
        backup_log, logger.main_log = logger.main_log, log_queue

        try:
            logger.info("{{c}}")
            logger.error("not expected")
        except Exception as e:
            self.assertTrue("c local is not found" in e)

    def test_warning_keyword_parameters(self):
        a = {"c": "a", "b": "d"}
        b = {"c": "b"}
        params = {"a": a, "b": b}

        WARNING = "WARNING: test"
        CAUSE = "\ncaused by\n\tERROR: Exception: problem"
        A = '{\n    "b": "d",\n    "c": "a"\n}'
        B = '{"c": "b"}'
        AC = "a"
        AB = "d"
        BC = "b"

        log_queue = StructuredLogger_usingQueue("abba")
        backup_log, logger.main_log = logger.main_log, log_queue

        try:
            raise Exception("problem")
        except Exception as e:
            logger.warning("test")
            self.assertEqual(logger.main_log.pop(), WARNING)

            logger.warning("test: {{a}}", a=a)
            self.assertEqual(logger.main_log.pop(), WARNING + ": " + A)

            logger.warning("test: {{a}}: {{b}}", a=a, b=b)
            self.assertEqual(logger.main_log.pop(), WARNING + ": " + A + ": " + B)

            logger.warning("test", e)
            log_value = logger.main_log.pop()
            self.assertEqual(log_value, WARNING + CAUSE)

            logger.warning("test: {{a}}", a=a, cause=e)
            self.assertEqual(logger.main_log.pop(), WARNING + ": " + A + CAUSE)

            logger.warning("test: {{a}}: {{b}}", a=a, b=b, cause=e)
            self.assertEqual(logger.main_log.pop(), WARNING + ": " + A + ": " + B + CAUSE)

            logger.warning("test: {{a}}", e, a=a)
            self.assertEqual(logger.main_log.pop(), WARNING + ": " + A + CAUSE)

            logger.warning("test: {{a}}: {{b}}", e, a=a, b=b)
            self.assertEqual(logger.main_log.pop(), WARNING + ": " + A + ": " + B + CAUSE)

            logger.warning("test: {{a}}", params, e)
            self.assertEqual(logger.main_log.pop(), WARNING + ": " + A + CAUSE)

            logger.warning("test: {{a}}: {{b}}", params, e)
            self.assertEqual(logger.main_log.pop(), WARNING + ": " + A + ": " + B + CAUSE)

            logger.warning("test: {{a}}: {{b}}", params, e, a=b)
            self.assertEqual(logger.main_log.pop(), WARNING + ": " + B + ": " + B + CAUSE)

            logger.warning("test: {{a}}: {{b}}", wrap(params), e, a=b)
            self.assertEqual(logger.main_log.pop(), WARNING + ": " + B + ": " + B + CAUSE)

            logger.warning("test: {{a.c}}", a=a)
            self.assertEqual(logger.main_log.pop(), WARNING + ": " + AC)

            logger.warning("test: {{a.c}}: {{a.b}}", a=a, b=b)
            self.assertEqual(logger.main_log.pop(), WARNING + ": " + AC + ": " + AB)

            logger.warning("test: {{a.c}}: {{b.c}}", a=a, b=b)
            self.assertEqual(logger.main_log.pop(), WARNING + ": " + AC + ": " + BC)

            logger.warning("test: {{a.c}}", a=a, cause=e)
            self.assertEqual(logger.main_log.pop(), WARNING + ": " + AC + CAUSE)

            logger.warning("test: {{a.c}}: {{b.c}}", a=a, b=b, cause=e)
            self.assertEqual(logger.main_log.pop(), WARNING + ": " + AC + ": " + BC + CAUSE)
        finally:
            logger.main_log = backup_log

    def test_note_keyword_parameters(self):
        a = {"c": "a", "b": "d"}
        b = {"c": "b"}
        params = {"a": a, "b": b}

        WARNING = "test"
        A = '{\n    "b": "d",\n    "c": "a"\n}'
        B = '{"c": "b"}'
        AC = "a"
        AB = "d"
        BC = "b"

        # DURING TESTING SOME OTHER THREADS MAY STILL BE WRITING TO THE LOG
        Till(seconds=1).wait()
        # HIGHJACK LOG FOR TESTING OUTPUT
        log_queue = StructuredLogger_usingQueue()
        backup_log, logger.main_log = logger.main_log, log_queue

        try:
            raise Exception("problem")
        except Exception as e:
            logger.info("test")
            self.assertEqual(logger.main_log.pop(), WARNING)

            logger.info("test: {{a}}", a=a)
            self.assertEqual(logger.main_log.pop(), WARNING + ": " + A)

            logger.info("test: {{a}}: {{b}}", a=a, b=b)
            self.assertEqual(logger.main_log.pop(), WARNING + ": " + A + ": " + B)

            logger.info("test: {{a.c}}", a=a)
            self.assertEqual(logger.main_log.pop(), WARNING + ": " + AC)

            logger.info("test: {{a}}: {{b}}", params)
            self.assertEqual(logger.main_log.pop(), WARNING + ": " + A + ": " + B)

            logger.info("test: {{a}}: {{b}}", params, a=b)
            self.assertEqual(logger.main_log.pop(), WARNING + ": " + B + ": " + B)

            logger.info("test: {{a}}: {{b}}", wrap(params), a=b)
            self.assertEqual(logger.main_log.pop(), WARNING + ": " + B + ": " + B)

            logger.info("test: {{a.c}}: {{a.b}}", a=a, b=b)
            self.assertEqual(logger.main_log.pop(), WARNING + ": " + AC + ": " + AB)

            logger.info("test: {{a.c}}: {{b.c}}", a=a, b=b)
            self.assertEqual(logger.main_log.pop(), WARNING + ": " + AC + ": " + BC)
        finally:
            logger.main_log = backup_log

    # NORMAL RAISING
    def test_python_raise_from(self):
        def problem_y():
            raise Exception("this is the root cause")

        def problem_x():
            try:
                problem_y()
            except Exception as e:
                raise Exception("this is a problem")

        logging.shutdown()
        from importlib import reload

        reload(logging)

        try:
            problem_x()
        except Exception as e:

            class _catcher(logging.Handler):
                def handle(self, record):
                    o = value2json(DataObject(record))
                    if record:
                        pass
                    if "this is a problem" not in e.args:
                        logger.error("We expect Python to, at least, report the first order problem")
                    if "this is the root cause" in e.args:
                        logger.error("We do not expect Python to report exception chains")

            log = logging.getLogger()
            log.addHandler(_catcher())
            log.exception("problem")

    # NORMAL RE-RAISE
    def test_python_re_raise(self):
        def problem_y():
            raise Exception("this is the root cause")

        def problem_x():
            try:
                problem_y()
            except Exception as f:
                raise f

        try:
            problem_x()
        except Exception as e:
            e = Except.wrap(e)
            self.assertEqual(e.cause, None)  # REALLY, THE CAUSE IS problem_y()

    def test_contains_from_zip_error(self):
        def bad_unzip():
            decompressor = zlib.decompressobj(16 + zlib.MAX_WBITS)
            return decompressor.decompress(b"invlaid zip file")

        try:
            bad_unzip()
            assert False
        except Exception as e:
            e = Except.wrap(e)
            if "incorrect header check" in e:
                pass
            else:
                assert False

    @skip("this is too complicated for now")
    def test_recursive_loop(self):
        def oh_no():
            try:
                oh_no()
            except BaseException as e:
                logger.error("this is a problem", e)

        try:
            oh_no()
            self.assertTrue(False, "should not happen")
        except Exception as e:
            self.assertIn("recursive", e, "expecting the recursive loop to be identified")

    @skip("this is too complicated for now")
    def test_deep_recursive_loop(self):
        def oh_no():
            try:
                fine1()
            except Exception as e:
                logger.error("this is a problem", e)

        def fine1():
            fine2()

        def fine2():
            oh_no()

        try:
            oh_no()
            self.assertTrue(False, "should not happen")
        except Exception as cause:
            self.assertIn("recursive", cause, "expecting the recursive loop to be identified")

    def test_locals_in_stack_trace(self):
        try:
            problem_c("test_value")
        except Exception as e:
            tb = sys.exc_info()[2]
            self.assertEqual(tb.tb_next.tb_frame.f_locals["a"].value, "test_value")

    def test_many_causes(self):
        try:
            logger.error("problem", cause=to_data([None]))
        except Exception as cause:
            self.assertEqual(cause.message, "problem")
            self.assertIsNone(cause.cause)

    def test_cause_depth_too_deep(self):
        cause = Except(ERROR, "not shown")
        for i in range(TOO_DEEP):
            cause = Except(ERROR, "error", cause=cause)
        text = str(cause)
        self.assertIn("not shown", text)

        cause = Except(ERROR, "error", cause=cause)
        text = str(cause)
        self.assertNotIn("not shown", text)


def problem_a():
    problem_b()


def problem_b():
    raise Exception("expected exception")


def problem_a2():
    try:
        problem_b()
    except Exception as e:
        logger.error("this is a problem", e)


def problem_c(value):
    a = Data(value=value)
    b = "something"
    c = 1 / 0


if __name__ == "__main__":
    try:
        Log.start()
        unittest.main()
    finally:
        Log.stop()



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from mo_future import allocate_lock

from mo_logs.log_usingNothing import StructuredLogger
from mo_logs.strings import expand_template


class StructuredLogger_usingPrint(StructuredLogger):
    def __init__(self):
        self.locker = allocate_lock()

    def write(self, template, params):
        value = expand_template(template, params)
        with self.locker:
            print(value)



# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Author: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from mo_testing.fuzzytestcase import FuzzyTestCase
from mo_times import Date

from mo_logs import strings
from mo_logs.strings import expand_template, wordify, round, datetime, parse_template


class TestStrings(FuzzyTestCase):
    def setUp(self):
        pass

    def test_right_align(self):
        total = 123.45
        some_list = [10, 11, 14, 80]

        result = expand_template("it is currently {{now|datetime}}", {"now": 1420119241000})
        self.assertEqual(result, "it is currently 2015-01-01 13:34:01")

        result = expand_template("Total: {{total|right_align(20)}}", {"total": total})
        self.assertEqual(result, "Total:               123.45")

        result = expand_template("Summary:\n{{list|json|indent}}", {"list": some_list})
        self.assertEqual(result, "Summary:\n\t[10, 11, 14, 80]")

        result = expand_template("Summary:\n{{list|indent}}", {"list": some_list})
        self.assertEqual(result, "Summary:\n\t[10, 11, 14, 80]")

    def test_no_align(self):
        details = {"person": {"name": "Kyle Lahnakoski", "age": 40}}
        result = expand_template("{{person.name}} is {{person.age}} years old", details)
        self.assertEqual(result, "Kyle Lahnakoski is 40 years old")

    def test_percent(self):
        self.assertEqual(strings.percent(0.123, digits=1), "10%")
        self.assertEqual(strings.percent(0.123, digits=2), "12%")
        self.assertEqual(strings.percent(0.123, digits=3), "12.3%")
        self.assertEqual(strings.percent(0.120, digits=3), "12.0%")

        self.assertEqual(strings.percent(0.0123, digits=1), "1%")
        self.assertEqual(strings.percent(0.0123, digits=2), "1.2%")
        self.assertEqual(strings.percent(0.0123, digits=3), "1.23%")
        self.assertEqual(strings.percent(0.0120, digits=3), "1.20%")

        self.assertEqual(strings.percent(0.5), "50%")

    def test_wordify(self):
        self.assertEqual(wordify("thisIsATest"), ["this", "is", "a", "test"])
        self.assertEqual(wordify("another.test"), ["another", "test"])
        self.assertEqual(wordify("also-a_test999"), ["also", "a", "test999"])
        self.assertEqual(wordify("BIG_WORDS"), ["big", "words"])
        self.assertEqual(wordify("ALSO_A_TEST999"), ["also", "a", "test999"])
        self.assertEqual(wordify("c:123:a"), ["c", "123", "a"])
        self.assertEqual(wordify("__int__"), ["__int__"])
        self.assertEqual(wordify(":"), [":"])
        self.assertEqual(wordify("__ENV__"), ["__env__"])

    def test_round(self):
        self.assertEqual(round(3.14), "3")

    def test_datatime(self):
        time = Date("2022-03-12")
        self.assertEqual(datetime(time), "2022-03-12 00:00:00")

    def test_quote(self):
        def f():
            return 1

        self.assertTrue(strings.quote(f).startswith('"<function TestStrings.test_quote.<locals>.f at'))

    def test_capitalize(self):
        result = expand_template("{{name|capitalize}}", {"name": "lahnakoski"})
        self.assertEqual(result, "Lahnakoski")

    def test_parse1(self):
        result = parse_template("{{name|capitalize}}")
        expected = [("", "name|capitalize")]
        self.assertEqual(result, expected)

    def test_parse2(self):
        result = parse_template("{{name|capitalize}} {{age}}")
        expected = [("", "name|capitalize"), (" ", "age")]
        self.assertEqual(result, expected)

    def test_parse3(self):
        result = parse_template("this is a test of {name}")
        expected = [("this is a test of ", "name")]
        self.assertEqual(result, expected)

    def test_parse4(self):
        result = parse_template("this is a test of {name|capitalize(\"some value\", lambda x: {'x': x})}")
        expected = [("this is a test of ", 'name|capitalize("some value", lambda x: {\'x\': x})')]
        self.assertEqual(result, expected)

    def test_parse5(self):
        result = parse_template("this is a test of {{name|capitalize(\"some value\", lambda x: {'x': x})}}")
        expected = [("this is a test of ", 'name|capitalize("some value", lambda x: {\'x\': x})')]
        self.assertEqual(result, expected)

    def test_parse6(self):
        result = parse_template("this is a test of {name|capitalize(\"some () value\")}")
        expected = [("this is a test of ", 'name|capitalize("some () value")')]
        self.assertEqual(result, expected)

    def test_parse_extra_curly(self):
        with self.assertRaises(Exception):
            parse_template("this is a test of {name|capitalize{(\"some () value\"}")

    def test_double_braces(self):
        result = parse_template("this is a {{{test}}} of {name|capitalize('some () value')}")
        expected = [("this is a ", "{test}"), (" of ", 'name|capitalize(\'some () value\')')]
        self.assertEqual(result, expected)

    def test_code(self):
        result = parse_template('a = "{"\nb="}"\n')
        expected=[('a = "{"\nb="}"\n', '')]
        self.assertEqual(result, expected)

    def test_double_quote(self):
        result = parse_template(' - ""{location.file}:{location.line}"" -')
        expected = [(' - "', 'location.file'), (':', 'location.line'), ('" -', '')]
        self.assertEqual(result, expected)




# encoding: utf-8
#
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
import base64
from decimal import Decimal

from mo_dots import to_data
from mo_future import number_types, is_text
from mo_json import datetime2unix, value2json
from mo_kwargs import override

from mo_logs import logger, strings
from mo_logs.exceptions import ALARM, ERROR, NOTE, WARNING
from mo_logs.log_usingNothing import StructuredLogger

LOG_STRING_LENGTH = 2000


class StructuredLogger_usingMozLog(StructuredLogger):
    """
    WRITE TO MozLog STANDARD FORMAT
    https://wiki.mozilla.org/Firefox/Services/Logging
    """

    @override
    def __init__(self, stream, app_name):
        """
        :param stream: MozLog IS A JSON FORMAT, WHICH IS BYTES
        :param app_name: MozLog WOULD LIKE TO KNOW WHAT APP IS MAKING THESE LOGS
        """
        self.stream = stream
        self.app_name = app_name
        if not app_name:
            logger.error("mozlog expects an `app_name` in the config")
        if not logger.trace:
            logger.error("mozlog expects trace=True so it gets the information it requires")

    def write(self, template, params):
        output = {
            "Timestamp": (Decimal(datetime2unix(params.timestamp)) * Decimal(1e9)).to_integral_exact(),  # NANOSECONDS
            "Type": params.template,
            "Logger": params.machine.name,
            "Hostname": self.app_name,
            "EnvVersion": "2.0",
            "Severity": severity_map.get(params.severity, 3),  # https://en.wikipedia.org/wiki/Syslog#Severity_levels
            "Pid": params.machine.pid,
            "Fields": {k: strings.limit(_json_to_string(v), LOG_STRING_LENGTH) for k, v in to_data(params).leaves()},
        }
        self.stream.write(value2json(output).encode("utf8"))
        self.stream.write(b"\n")


def _json_to_string(value) -> str:
    """
    :param value: SOME STRUCTURE
    :return: str
    """
    if isinstance(value, number_types):
        return str(value)
    elif is_text(value):
        return value
    elif isinstance(value, bytes):
        return base64.b64encode(value).decode("latin1")
    else:
        return value2json(value)


# https://en.wikipedia.org/wiki/Syslog#Severity_levels
severity_map = {
    ERROR: 3,  # Error
    WARNING: 4,  # Warning
    ALARM: 5,  # Notice
    NOTE: 6,  # Informational
}


def datatime2decimal(value):
    return
